"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _three = require("three");

var _Coordinates = _interopRequireDefault(require("./Coordinates"));

var _Extent = _interopRequireDefault(require("./Extent"));

/**
 * Generated On: 2015-10-5
 * Class: Projection
 * Description: Outils de projections cartographiques et de convertion
 */
var PI_OV_FOUR = Math.PI / 4;
var PI_OV_TWO = Math.PI / 2;
var INV_TWO_PI = 1.0 / (Math.PI * 2);
var LOG_TWO = Math.log(2.0); // TODO: Clamp to 85.0511288° because:
// longitude 180° to X EPSG:3857 = 20037508.34m
// To get a square, we convert Y 20037508.34m to latitude, we get 85.0511288°
// Change it when we will change worldDimension3857 in Extent.js

function WGS84LatitudeClamp(latitude) {
  return Math.min(84, Math.max(-86, latitude));
}

var dim = {
  x: 0,
  y: 0
};
var center = new _Coordinates["default"]('EPSG:4326', 0, 0, 0);
var Projection = {
  /**
   * Convert latitude to y coordinate in pseudo mercator (EPSG:3857)
   * @param {number} latitude - latitude in degrees (EPSG:4326)
   * @return {number} y coordinate in pseudo mercator
   */
  latitudeToY_PM: function latitudeToY_PM(latitude) {
    return 0.5 - Math.log(Math.tan(PI_OV_FOUR + _three.Math.degToRad(latitude) * 0.5)) * INV_TWO_PI;
  },

  /**
   * Convert from y coordinate pseudo mercator (EPSG:3857) to latitude
   * @param {number} y - y coordinate in pseudo mercator
   * @return {number} - latitude in degrees (EPSG:4326)
   */
  y_PMTolatitude: function y_PMTolatitude(y) {
    return _three.Math.radToDeg(2 * (Math.atan(Math.exp(-(y - 0.5) / INV_TWO_PI)) - PI_OV_FOUR));
  },
  computeWmtsPm: function computeWmtsPm(extent_wmtsWgs84g, extent_epsg4326) {
    var extents_WMTS_PM = [];
    var level = extent_wmtsWgs84g.zoom + 1;
    var nbRow = Math.pow(2, level);
    var sizeRow = 1.0 / nbRow;
    var yMin = Projection.latitudeToY_PM(WGS84LatitudeClamp(extent_epsg4326.north));
    var yMax = Projection.latitudeToY_PM(WGS84LatitudeClamp(extent_epsg4326.south));
    var maxRow;
    var minRow = Math.floor(yMin / sizeRow); // ]N; N+1] => N

    maxRow = Math.ceil(yMax / sizeRow) - 1; // make sure we don't exceed boundaries

    maxRow = Math.min(maxRow, nbRow - 1);
    var minCol = extent_wmtsWgs84g.col;

    for (var r = maxRow; r >= minRow; r--) {
      for (var c = minCol; c <= minCol; c++) {
        extents_WMTS_PM.push(new _Extent["default"]('WMTS:PM', level, r, c));
      }
    }

    return extents_WMTS_PM;
  },
  extent_Epsg4326_To_WmtsWgs84g: function extent_Epsg4326_To_WmtsWgs84g(extent_epsg4326) {
    var extent_wmtsWgs84g = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : new _Extent["default"]('WMTS:WGS84G', 0, 0, 0);
    extent_epsg4326.dimensions(dim);
    var zoom = Math.floor(Math.log(Math.PI / _three.Math.degToRad(dim.y)) / LOG_TWO + 0.5);
    var nY = Math.pow(2, zoom);
    var uX = Math.PI * 2 / (2 * nY);
    var uY = Math.PI / nY;
    extent_epsg4326.center(center);
    var col = Math.floor((Math.PI + _three.Math.degToRad(center.longitude)) / uX);
    var row = Math.floor(nY - (PI_OV_TWO + _three.Math.degToRad(center.latitude)) / uY);
    return extent_wmtsWgs84g.set(zoom, row, col);
  }
};
var _default = Projection;
exports["default"] = _default;